@@ -1,5 +1,5 @@
 /* Data base of default implicit rules for GNU Make.
-Copyright (C) 1988, 1989, 1990, 1991, 1992 Free Software Foundation, Inc.
+Copyright (C) 1988, 1989, 1990, 1991, 1992, 1993 Free Software Foundation, Inc.
 This file is part of GNU Make.
 
 GNU Make is free software; you can redistribute it and/or modify
@@ -30,7 +30,8 @@ the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */
 
 static char default_suffixes[]
   = ".out .a .ln .o .c .cc .C .p .f .F .r .y .l .s .S \
-.mod .sym .def .h .info .dvi .tex .texinfo .texi .cweb .web .sh .elc .el";
+.mod .sym .def .h .info .dvi .tex .texinfo .texi .txinfo \
+.cweb .web .sh .elc .el";
 
 static struct pspec default_pattern_rules[] =
   {
@@ -150,10 +151,13 @@ static char *default_suffix_rules[] =
     "$(PREPROCESS.S) $< > $@",
 
     ".texinfo.info",
-    "$(MAKEINFO) $<",
+    "$(MAKEINFO) $< -o $@",
 
     ".texi.info",
-    "$(MAKEINFO) $<",
+    "$(MAKEINFO) $< -o $@",
+
+    ".txinfo.info",
+    "$(MAKEINFO) $< -o $@",
 
     ".tex.dvi",
     "$(TEX) $<",
@@ -164,6 +168,9 @@ static char *default_suffix_rules[] =
     ".texi.dvi",
     "$(TEXI2DVI) $<",
 
+    ".txinfo.dvi",
+    "$(TEXI2DVI) $<",
+
     ".cweb.c",
     "$(CTANGLE) $<",
 
@@ -216,11 +223,7 @@ static char *default_variables[] =
     "F77", "$(FC)",
     "F77FLAGS", "$(FFLAGS)",
 #endif	/* Cray.  */
-#if	defined (USG) || defined (ultrix) || defined (_IBMR2)
-    "GET", "get",
-#else
-    "GET", "/usr/sccs/get",
-#endif
+    "GET", SCCS_GET,
     "LD", "ld",
     "LEX", "lex",
     "LINT", "lint",
@@ -303,7 +306,7 @@ set_default_suffixes ()
       char *p = default_suffixes;
       suffix_file->deps = (struct dep *)
 	multi_glob (parse_file_seq (&p, '\0', sizeof (struct dep)),
-		    sizeof (struct dep));
+		    sizeof (struct dep), 1);
       (void) define_variable ("SUFFIXES", 8, default_suffixes, o_default, 0);
     }
 }
