@@ -1141,8 +1141,12 @@ string_glob (line)
   register struct nameseq *chain;
   register unsigned int idx;
 
-  chain = multi_glob (parse_file_seq (&line, '\0', sizeof (struct nameseq)),
-		      sizeof (struct nameseq));
+  chain = multi_glob (parse_file_seq (&line, '\0', sizeof (struct nameseq), 0),
+		      sizeof (struct nameseq),
+		      /* We do not want parse_file_seq to strip `./'s.
+			 That would break examples like:
+			   $(patsubst ./%.c,obj/%.o,$(wildcard ./*.c))  */
+		      0);
 
   if (result == 0)
     {
