@@ -344,9 +344,6 @@ reap_children (block, err)
 		  if (c->file->update_status != 0)
 		    /* We failed to start the commands.  */
 		    delete_child_targets (c);
-
-		  /* Tell update_file that some actual work has been done.  */
-		  ++files_remade;
 		  break;
 
 		default:
@@ -415,20 +412,13 @@ free_child (child)
 #ifdef	POSIX
 extern sigset_t fatal_signal_set;
 
-static void
+void
 unblock_sigs ()
 {
   sigset_t empty;
   sigemptyset (&empty);
   sigprocmask (SIG_SETMASK, &empty, (sigset_t *) 0);
 }
-#else
-#ifdef	HAVE_SIGSETMASK
-extern int fatal_signal_mask;
-#define	unblock_sigs()	sigsetmask (0)
-#else
-#define	unblock_sigs()
-#endif
 #endif
 
 /* Start a job to run the commands specified in CHILD.
@@ -440,17 +430,15 @@ start_job_command (child)
 {
   static int bad_stdin = -1;
   register char *p;
-  char noprint = 0, recursive;
+  int flags = child->file->cmds->lines_flags[child->command_line - 1];
   char **argv;
 
-  recursive = child->file->cmds->lines_recurse[child->command_line - 1];
-
   p = child->command_ptr;
-  child->noerror = 0;
+  child->noerror = flags & COMMANDS_NOERROR;
   while (*p != '\0')
     {
       if (*p == '@')
-	noprint = 1;
+	flags |= COMMANDS_SILENT;
       else if (*p == '-')
 	child->noerror = 1;
       else if (!isblank (*p) && *p != '+')
@@ -459,7 +447,7 @@ start_job_command (child)
     }
 
   /* If -q was given, just say that updating `failed'.  */
-  if (question_flag && !recursive)
+  if (question_flag && !(flags & COMMANDS_RECURSE))
     goto error;
 
   /* There may be some preceding whitespace left if there
@@ -480,7 +468,7 @@ start_job_command (child)
       }
   }
 
-  if (touch_flag && !recursive)
+  if (touch_flag && !(flags & COMMANDS_RECURSE))
     {
       /* Go on to the next command.  It might be the recursive one.
 	 We construct ARGV only to find the end of the command line.  */
@@ -499,22 +487,26 @@ start_job_command (child)
 
   /* Print out the command.  */
 
-  if (just_print_flag || (!noprint && !silent_flag))
+  if (just_print_flag || (!(flags & COMMANDS_SILENT) && !silent_flag))
     puts (p);
 
+  /* Tell update_goal_chain that a command has been started on behalf of
+     this target.  It is important that this happens here and not in
+     reap_children (where we used to do it), because reap_children might be
+     reaping children from a different target.  We want this increment to
+     guaranteedly indicate that a command was started for the dependency
+     chain (i.e., update_file recursion chain) we are processing.  */
+
+  ++commands_started;
+
   /* If -n was given, recurse to get the next line in the sequence.  */
 
-  if (just_print_flag && !recursive)
+  if (just_print_flag && !(flags & COMMANDS_RECURSE))
     {
       free (argv[0]);
       free ((char *) argv);
       if (job_next_command (child))
 	start_job_command (child);
-      else
-	/* Normally, this is set by reap_children to indicate that
-	   some commands were actually run.  Under -n, reap_children
-	   never gets called, so we increment it here.  */
-	++files_remade;
       return;
     }
 
@@ -619,7 +611,11 @@ start_job_command (child)
   child->file->command_state = cs_finished;
 }
 
-static void
+/* Try to start a child running.
+   Returns nonzero if the child was started (and maybe finished), or zero if
+   the load was too high and the child was put on the `waiting_jobs' chain.  */
+
+static int
 start_waiting_job (c)
      struct child *c;
 {
@@ -638,7 +634,7 @@ start_waiting_job (c)
       c->file->command_state = cs_running;
       c->next = waiting_jobs;
       waiting_jobs = c;
-      return;
+      return 0;
     }
 
   /* Start the first command; reap_children will run later command lines.  */
@@ -669,6 +665,8 @@ start_waiting_job (c)
       abort ();
       break;
     }
+
+  return 1;
 }
 
 /* Create a `struct child' for FILE and start its commands running.  */
@@ -682,6 +680,10 @@ new_job (file)
   char **lines;
   register unsigned int i;
 
+  /* Let any previously decided-upon jobs that are waiting
+     for the load to go down start before this new one.  */
+  start_waiting_jobs ();
+
   /* Reap any children that might have finished recently.  */
   reap_children (0, 0);
 
@@ -696,8 +698,103 @@ new_job (file)
   /* Expand the command lines and store the results in LINES.  */
   lines = (char **) xmalloc (cmds->ncommand_lines * sizeof (char *));
   for (i = 0; i < cmds->ncommand_lines; ++i)
-    lines[i] = allocated_variable_expand_for_file (cmds->command_lines[i],
-						   file);
+    {
+      /* Collapse backslash-newline combinations that are inside variable
+	 or function references.  These are left alone by the parser so
+	 that they will appear in the echoing of commands (where they look
+	 nice); and collapsed by construct_command_argv when it tokenizes.
+	 But letting them survive inside function invocations loses because
+	 we don't want the functions to see them as part of the text.  */
+
+      char *in, *out, *ref;
+
+      /* IN points to where in the line we are scanning.
+	 OUT points to where in the line we are writing.
+	 When we collapse a backslash-newline combination,
+	 IN gets ahead out OUT.  */
+
+      in = out = cmds->command_lines[i];
+      while ((ref = index (in, '$')) != 0)
+	{
+	  ++ref;		/* Move past the $.  */
+
+	  if (out != in)
+	    /* Copy the text between the end of the last chunk
+	       we processed (where IN points) and the new chunk
+	       we are about to process (where REF points).  */
+	    bcopy (in, out, ref - in);
+
+	  /* Move both pointers past the boring stuff.  */
+	  out += ref - in;
+	  in = ref;
+
+	  if (*ref == '(' || *ref == '{')
+	    {
+	      char openparen = *ref;
+	      char closeparen = openparen == '(' ? ')' : '}';
+	      int count;
+	      char *p;
+
+	      *out++ = *in++;	/* Copy OPENPAREN.  */
+	      /* IN now points past the opening paren or brace.
+		 Count parens or braces until it is matched.  */
+	      count = 0;
+	      while (*in != '\0')
+		{
+		  if (*in == closeparen && --count < 0)
+		    break;
+		  else if (*in == '\\' && in[1] == '\n')
+		    {
+		      /* We have found a backslash-newline inside a
+			 variable or function reference.  Eat it and
+			 any following whitespace.  */
+
+		      int quoted = 0;
+		      for (p = in - 1; p > ref && *p == '\\'; --p)
+			quoted = !quoted;
+
+		      if (quoted)
+			/* There were two or more backslashes, so this is
+			   not really a continuation line.  We don't collapse
+			   the quoting backslashes here as is done in
+			   collapse_continuations, because the line will
+			   be collapsed again after expansion.  */
+			*out++ = *in++;
+		      else
+			{
+			  /* Skip the backslash, newline and
+			     any following whitespace.  */
+			  in = next_token (in + 2);
+
+			  /* Discard any preceding whitespace that has
+			     already been written to the output.  */
+			  while (out > ref && isblank (out[-1]))
+			    --out;
+
+			  /* Replace it all with a single space.  */
+			  *out++ = ' ';
+			}
+		    }
+		  else
+		    {
+		      if (*in == openparen)
+			++count;
+
+		      *out++ = *in++;
+		    }
+		}
+	    }
+	}
+
+      /* There are no more references in this line to worry about.
+	 Copy the remaining uninteresting text to the output.  */
+      if (out != in)
+	strcpy (out, in);
+
+      /* Finally, expand the line.  */
+      lines[i] = allocated_variable_expand_for_file (cmds->command_lines[i],
+						     file);
+    }
 
   /* Start the command sequence, record it in a new
      `struct child', and add that to the chain.  */
@@ -786,22 +883,23 @@ load_too_high ()
 void
 start_waiting_jobs ()
 {
-  while (waiting_jobs != 0)
-    {
-      struct child *c;
+  struct child *job;
 
+  if (waiting_jobs == 0)
+    return;
+
+  do
+    {
       /* Check for recently deceased descendants.  */
       reap_children (0, 0);
 
-      if (job_slots_used > 0
-	  && (job_slots_used == job_slots || load_too_high ()))
-	/* We have started all the jobs we can at the moment.  */
-	return;
+      /* Take a job off the waiting list.  */
+      job = waiting_jobs;
+      waiting_jobs = job->next;
 
-      c = waiting_jobs;
-      waiting_jobs = c->next;
-      start_waiting_job (c);
-    }
+      /* Try to start that job.  We break out of the loop as soon
+	 as start_waiting_job puts one back on the waiting list.  */
+    } while (start_waiting_job (job) && waiting_jobs != 0);
 }
 
 /* Replace the current process with one executing the command in ARGV.
