@@ -83,6 +83,7 @@ flags () {}
 struct command_switch
   {
     char c;			/* The switch character.  */
+
     enum			/* Type of the value.  */
       {
 	flag,			/* Turn int flag on.  */
@@ -168,11 +169,15 @@ int no_builtin_rules_flag = 0;
 int keep_going_flag;
 int default_keep_going_flag = 0;
 
-/* Nonzero means print working directory
-   before starting and after finishing make. */
+/* Nonzero means print directory before starting and when done (-w).  */
 
 int print_directory_flag = 0;
 
+/* Nonzero means ignore print_directory_flag and never print the directory.
+   This is necessary because print_directory_flag is set implicitly.  */
+
+int inhibit_print_directory_flag = 0;
+
 /* Nonzero means print version information.  */
 
 int print_version_flag = 0;
@@ -217,7 +222,7 @@ static struct stringlist *new_files = 0;
 
 /* The table of command switches.  */
 
-static struct command_switch switches[] =
+static const struct command_switch switches[] =
   {
     { 'b', ignore, 0, 0, 0, 0, 0, 0,
 	0, 0,
@@ -289,6 +294,9 @@ static struct command_switch switches[] =
     { 'w', flag, (char *) &print_directory_flag, 1, 1, 0, 0, 0,
 	"print-directory", 0,
 	"Print the current directory" },
+    { 1, flag, (char *) &inhibit_print_directory_flag, 1, 1, 0, 0, 0,
+	"no-print-directory", 0,
+	"Turn off -w, even if it was turned on implicitly" },
     { 'W', string, (char *) &new_files, 0, 0, 0, 0, 0,
 	"what-if", "FILE",
 	"Consider FILE to be infinitely new" },
@@ -301,7 +309,6 @@ static struct option long_option_aliases[] =
   {
     { "quiet",		no_argument,		0, 's' },
     { "stop",		no_argument,		0, 'S' },
-    { "new",		required_argument,	0, 'W' },
     { "new-file",	required_argument,	0, 'W' },
     { "assume-new",	required_argument,	0, 'W' },
     { "assume-old",	required_argument,	0, 'o' },
@@ -325,6 +332,10 @@ struct stringlist *other_args = 0;
 
 char *program;
 
+/* Our current directory after processing all -C options.  */
+
+char *starting_directory;
+
 /* Value of the MAKELEVEL variable at startup (or 0).  */
 
 unsigned int makelevel;
@@ -369,6 +380,7 @@ main (argc, argv, envp)
   register struct dep *lastgoal;
   struct dep *read_makefiles;
   PATH_VAR (current_directory);
+  char *directory_before_chdir;
 
   default_goal_file = 0;
   reading_filename = 0;
@@ -457,7 +469,11 @@ main (argc, argv, envp)
       error ("getwd: %s", current_directory);
 #endif
       current_directory[0] = '\0';
+      directory_before_chdir = 0;
     }
+  else
+    directory_before_chdir = savestring (current_directory,
+					 strlen (current_directory));
 
   /* Read in variables from the environment.  It is important that this be
      done before `MAKE' and `MAKEOVERRIDES' are figured out so their
@@ -498,7 +514,8 @@ main (argc, argv, envp)
 	continue;
 
       /* Try a variable definition.  */
-      if (try_variable_definition (other_args->list[i], o_command))
+      if (try_variable_definition ((char *) 0, 0,
+				   other_args->list[i], o_command))
 	{
 	  /* It succeeded.  The variable is already defined.
 	     Backslash-quotify it and append it to CMD_DEFS, then clobber it
@@ -592,11 +609,34 @@ main (argc, argv, envp)
   if (!silent_flag && (directories != 0 || makelevel > 0))
     print_directory_flag = 1;
 
+  /* Let the user disable that with --no-print-directory.  */
+  if (inhibit_print_directory_flag)
+    print_directory_flag = 0;
+
   /* Construct the list of include directories to search.  */
 
   construct_include_path (include_directories == 0 ? (char **) 0
 			  : include_directories->list);
 
+  /* Figure out where we are now, after chdir'ing.  */
+  if (directories == 0)
+    /* We didn't move, so we're still in the same place.  */
+    starting_directory = current_directory;
+  else
+    {
+      if (getcwd (current_directory, GET_PATH_MAX) == 0)
+	{
+#ifdef	HAVE_GETCWD
+	  perror_with_name ("getcwd: ", "");
+#else
+	  error ("getwd: %s", current_directory);
+#endif
+	  starting_directory = 0;
+	}
+      else
+	starting_directory = current_directory;
+    }
+
   /* Tell the user where he is.  */
 
   if (print_directory_flag)
@@ -656,9 +696,10 @@ main (argc, argv, envp)
   /* Set up to handle children dying.  This must be done before
      reading in the makefiles so that `shell' function calls will work.  */
 
-#if	defined (SIGCHLD) && !defined (USG)
+#ifdef SIGCHLD
   (void) signal (SIGCHLD, child_handler);
-#else
+#endif
+#ifdef SIGCLD
   (void) signal (SIGCLD, child_handler);
 #endif
 
@@ -903,21 +944,21 @@ main (argc, argv, envp)
 	      register unsigned int i, j = 0;
 	      for (i = 1; i < argc; ++i)
 		if (!strcmp (argv[i], "-f"))
-		    {
-		      char *p = &argv[i][2];
-		      if (*p == '\0')
-			argv[++i] = makefiles->list[j];
-		      else
-			argv[i] = concat ("-f", makefiles->list[j], "");
-		      ++j;
-		    }
+		  {
+		    char *p = &argv[i][2];
+		    if (*p == '\0')
+		      argv[++i] = makefiles->list[j];
+		    else
+		      argv[i] = concat ("-f", makefiles->list[j], "");
+		    ++j;
+		  }
 	    }
 	  if (directories != 0 && directories->idx > 0)
 	    {
 	      char bad;
-	      if (current_directory[0] != '\0')
+	      if (directory_before_chdir != 0)
 		{
-		  if (chdir (current_directory) < 0)
+		  if (chdir (directory_before_chdir) < 0)
 		    {
 		      perror_with_name ("chdir", "");
 		      bad = 1;
@@ -992,8 +1033,7 @@ main (argc, argv, envp)
 	    abort ();
 	  }
       }
-
-    if (goals == 0)
+    else
       {
 	if (read_makefiles == 0)
 	  fatal ("No targets specified and no makefile found");
@@ -1033,7 +1073,9 @@ init_switches ()
       long_options[i].name = (switches[i].long_name == 0 ? "" :
 			      switches[i].long_name);
       long_options[i].flag = 0;
-      *p++ = long_options[i].val = switches[i].c;
+      long_options[i].val = switches[i].c;
+      if (isalnum (switches[i].c))
+	*p++ = switches[i].c;
       switch (switches[i].type)
 	{
 	case flag:
@@ -1046,10 +1088,12 @@ init_switches ()
 	case string:
 	case positive_int:
 	case floating:
-	  *p++ = ':';
+	  if (isalnum (switches[i].c))
+	    *p++ = ':';
 	  if (switches[i].noarg_value != 0)
 	    {
-	      *p++ = ':';
+	      if (isalnum (switches[i].c))
+		*p++ = ':';
 	      long_options[i].has_arg = optional_argument;
 	    }
 	  else
@@ -1075,7 +1119,7 @@ decode_switches (argc, argv, env)
      int env;
 {
   int bad = 0;
-  register struct command_switch *cs;
+  register const struct command_switch *cs;
   register struct stringlist *sl;
   register int c;
 
@@ -1252,12 +1296,18 @@ positive integral argument",
 	    }
 
 	  p = buf;
-	  sprintf (buf, "  -%c%s", cs->c, arg);
-	  p += strlen (p);
+
+	  if (isalnum (cs->c))
+	    {
+	      sprintf (buf, "  -%c%s", cs->c, arg);
+	      p += strlen (p);
+	    }
 	  if (cs->long_name != 0)
 	    {
 	      unsigned int i;
-	      sprintf (p, ", --%s%s", cs->long_name, arg);
+	      sprintf (p, "%s--%s%s",
+		       !isalnum (cs->c) ? "  " : ", ",
+		       cs->long_name, arg);
 	      p += strlen (p);
 	      for (i = 0; i < (sizeof (long_option_aliases) /
 			       sizeof (long_option_aliases[0]));
@@ -1269,7 +1319,7 @@ positive integral argument",
 		  }
 	    }
 	  {
-	    struct command_switch *ncs = cs;
+	    const struct command_switch *ncs = cs;
 	    while ((++ncs)->c != '\0')
 	      if (ncs->description[0] == '-' &&
 		  ncs->description[1] == cs->c)
@@ -1345,7 +1395,7 @@ decode_env_switches (envar, len)
   args[(value[0] == '-' ? 0 : 1) + len + 1] = '\0';
 
   /* Allocate a vector that is definitely big enough.  */
-  argv = (char **) alloca (1 + len * sizeof (char *));
+  argv = (char **) alloca ((1 + len + 1) * sizeof (char *));
 
   /* getopt will look at the arguments starting at ARGV[1].
      Prepend a spacer word.  */
@@ -1371,7 +1421,7 @@ static void
 define_makeflags (all, makefile)
      int all, makefile;
 {
-  register struct command_switch *cs;
+  register const struct command_switch *cs;
   char *flagstring;
 
   /* We will construct a linked list of `struct flag's describing
@@ -1382,7 +1432,7 @@ define_makeflags (all, makefile)
   struct flag
     {
       struct flag *next;
-      int c;
+      const struct command_switch *cs;
       char *arg;
       unsigned int arglen;
     };
@@ -1391,7 +1441,7 @@ define_makeflags (all, makefile)
 #define	ADD_FLAG(ARG, LEN) \
   do {									      \
     struct flag *new = (struct flag *) alloca (sizeof (struct flag));	      \
-    new->c = cs->c;							      \
+    new->cs = cs;							      \
     new->arg = (ARG);							      \
     new->arglen = (LEN);						      \
     new->next = flags;							      \
@@ -1400,6 +1450,9 @@ define_makeflags (all, makefile)
       ++flagslen;		/* Just a single flag letter.  */	      \
     else								      \
       flagslen += 1 + 1 + 1 + 1 + new->arglen; /* " -x foo" */		      \
+    if (!isalnum (cs->c))						      \
+      /* This switch has no single-letter version, so we use the long.  */    \
+      flagslen += 2 + strlen (cs->long_name);				      \
   } while (0)
 
   for (cs = switches; cs->c != '\0'; ++cs)
@@ -1430,7 +1483,7 @@ define_makeflags (all, makefile)
 	      else if (cs->noarg_value != 0
 		       && (*(unsigned int *) cs->value_ptr ==
 			   *(unsigned int *) cs->noarg_value))
-		ADD_FLAG (0, 0);
+		ADD_FLAG ("", 0); /* Optional value omitted; see below.  */
 	      else if (cs->c == 'j')
 		/* Special case for `-j'.  */
 		ADD_FLAG ("1", 1);
@@ -1453,7 +1506,7 @@ define_makeflags (all, makefile)
 	      else if (cs->noarg_value != 0
 		       && (*(double *) cs->value_ptr
 			   == *(double *) cs->noarg_value))
-		ADD_FLAG (0, 0);
+		ADD_FLAG ("", 0); /* Optional value omitted; see below.  */
 	      else
 		{
 		  char *buf = (char *) alloca (100);
@@ -1467,10 +1520,15 @@ define_makeflags (all, makefile)
 	  if (all)
 	    {
 	      struct stringlist *sl = *(struct stringlist **) cs->value_ptr;
-	      register unsigned int i;
 	      if (sl != 0)
-		for (i = 0; i < sl->idx; ++i)
-		  ADD_FLAG (sl->list[i], strlen (sl->list[i]));
+		{
+		  /* Add the elements in reverse order, because
+		     all the flags get reversed below; and the order
+		     matters for some switches (like -I).  */
+		  register unsigned int i = sl->idx;
+		  while (i-- > 0)
+		    ADD_FLAG (sl->list[i], strlen (sl->list[i]));
+		}
 	    }
 	  break;
 	}
@@ -1490,18 +1548,39 @@ define_makeflags (all, makefile)
       *p++ = '-';
       do
 	{
-	  /* Add the flag letter to the string.  */
-	  *p++ = flags->c;
+	  /* Add the flag letter or name to the string.  */
+	  if (!isalnum (flags->cs->c))
+	    {
+	      *p++ = '-';
+	      strcpy (p, flags->cs->long_name);
+	      p += strlen (p);
+	    }
+	  else
+	    *p++ = flags->cs->c;
 	  if (flags->arg != 0)
 	    {
-	      /* Add its argument too.  */
-	      *p++ = ' ';
-	      bcopy (flags->arg, p, flags->arglen);
-	      p += flags->arglen;
+	      /* A flag that takes an optional argument which in this case
+		 is omitted is specified by ARG being "" and ARGLEN being 0.
+		 We must distinguish because a following flag appended without
+		 an intervening " -" is considered the arg for the first.  */
+	      if (flags->arglen > 0)
+		{
+		  /* Add its argument too.  */
+		  *p++ = !isalnum (flags->cs->c) ? '=' : ' ';
+		  bcopy (flags->arg, p, flags->arglen);
+		  p += flags->arglen;
+		}
 	      /* Write a following space and dash, for the next flag.  */
 	      *p++ = ' ';
 	      *p++ = '-';
 	    }
+	  else if (!isalnum (flags->cs->c))
+	    {
+	      /* Long options must each go in their own word,
+		 so we write the following space and dash.  */
+	      *p++ = ' ';
+	      *p++ = '-';
+	    }
 	  flags = flags->next;
 	} while (flags != 0);
 
@@ -1613,7 +1692,6 @@ log_working_directory (entering)
      int entering;
 {
   static int entered = 0;
-  PATH_VAR (pwdbuf);
   char *message = entering ? "Entering" : "Leaving";
 
   if (entering)
@@ -1631,15 +1709,8 @@ log_working_directory (entering)
   else
     printf ("%s[%u]: %s ", program, makelevel, message);
 
-  if (getcwd (pwdbuf, GET_PATH_MAX) == 0)
-    {
-#ifdef	HAVE_GETCWD
-      perror_with_name ("getcwd: ", "");
-#else
-      error ("getwd: %s", pwdbuf);
-#endif
-      puts ("an unknown directory");
-    }
+  if (starting_directory == 0)
+    puts ("an unknown directory");
   else
-    printf ("directory `%s'\n", pwdbuf);
+    printf ("directory `%s'\n", starting_directory);
 }
