@@ -1,4 +1,5 @@
-/* Copyright (C) 1988, 1989, 1990, 1991, 1992, 1993 Free Software Foundation, Inc.
+/* Copyright (C) 1988, 1989, 1990, 1991, 1992, 1993
+	Free Software Foundation, Inc.
 This file is part of GNU Make.
 
 GNU Make is free software; you can redistribute it and/or modify
@@ -343,6 +344,8 @@ read_makefile (filename, type)
       p = collapsed;
       while (isspace (*p))
 	++p;
+      /* We cannot consider a line containing just a tab to be empty
+	 because it might constitute an empty command for a target.  */
       if (*p == '\0' && lb.buffer[0] != '\t')
 	continue;
 
@@ -401,7 +404,9 @@ read_makefile (filename, type)
 				      lineno, infile, filename);
 		}
 	    }
-	  else if (!ignoring && !try_variable_definition (p2, o_override))
+	  else if (!ignoring
+		   && !try_variable_definition (filename, lineno,
+						p2, o_override))
 	    makefile_error (filename, lineno, "empty `override' directive");
 
 	  continue;
@@ -418,39 +423,40 @@ read_makefile (filename, type)
 	  unsigned int len;
 
 	  if (no_targets)
+	    /* Ignore the commands in a rule with no targets.  */
+	    continue;
+
+	  /* If there is no preceding rule line, don't treat this line
+	     as a command, even though it begins with a tab character.
+	     SunOS 4 make appears to behave this way.  */
+
+	  if (filenames != 0)
 	    {
-	      /* Ignore the commands in a rule with no targets.  */
-	      no_targets = 0;
+	      /* Append this command line to the line being accumulated.  */
+	      p = lb.buffer;
+	      if (commands_idx == 0)
+		commands_started = lineno;
+	      len = strlen (p);
+	      if (len + 1 + commands_idx > commands_len)
+		{
+		  commands_len = (len + 1 + commands_idx) * 2;
+		  commands = (char *) xrealloc (commands, commands_len);
+		}
+	      bcopy (p, &commands[commands_idx], len);
+	      commands_idx += len;
+	      commands[commands_idx++] = '\n';
+
 	      continue;
 	    }
-
-	  if (filenames == 0)
-	    makefile_fatal (filename, lineno,
-			    "commands with no associated target");
-
-	  /* Add this command line to end of the line being accumulated.  */
-	  p = lb.buffer;
-	  if (commands_idx == 0)
-	    commands_started = lineno;
-	  len = strlen (p);
-	  if (len + 1 + commands_idx > commands_len)
-	    {
-	      commands_len = (len + 1 + commands_idx) * 2;
-	      commands = (char *) xrealloc (commands, commands_len);
-	    }
-	  bcopy (p, &commands[commands_idx], len);
-	  commands_idx += len;
-	  commands[commands_idx++] = '\n';
-
-	  continue;
 	}
-      else if (word1eq ("export", 6))
+
+      if (word1eq ("export", 6))
 	{
 	  struct variable *v;
 	  p2 = next_token (p + 6);
 	  if (*p2 == '\0')
 	    export_all_variables = 1;
-	  v = try_variable_definition (p2, o_file);
+	  v = try_variable_definition (filename, lineno, p2, o_file);
 	  if (v != 0)
 	    v->export = v_export;
 	  else
@@ -500,7 +506,8 @@ read_makefile (filename, type)
 	  p2 = p;
 	  files = multi_glob (parse_file_seq (&p2, '\0',
 					      sizeof (struct nameseq)),
-			      sizeof (struct nameseq));
+			      sizeof (struct nameseq),
+			      1);
 	  free (p);
 
 	  /* Save the state of conditionals and start
@@ -539,8 +546,6 @@ read_makefile (filename, type)
 	    {
 	      pattern = savestring (p, len);
 	      p = find_next_token (&p2, &len);
-	      if (p != 0)
-		p = savestring (p, len);
 	      /* No searchpath means remove all previous
 		 selective VPATH's with the same pattern.  */
 	    }
@@ -550,11 +555,9 @@ read_makefile (filename, type)
 	  construct_vpath_list (pattern, p);
 	  if (pattern != 0)
 	    free (pattern);
-	  if (p != 0)
-	    free (p);
 	}
 #undef	word1eq
-      else if (try_variable_definition (p, o_file))
+      else if (try_variable_definition (filename, lineno, p, o_file))
 	/* This line has been dealt with.  */
 	;
       else
@@ -610,8 +613,9 @@ read_makefile (filename, type)
 	    }
 
 	  filenames = multi_glob (parse_file_seq (&p2, ':',
-						  sizeof (struct nameseq), 1),
-				  sizeof (struct nameseq));
+						  sizeof (struct nameseq)),
+				  sizeof (struct nameseq),
+				  1);
 	  if (*p2++ == '\0')
 	    makefile_fatal (filename, lineno, "missing separator");
 	  /* Is this a one-colon or two-colon entry?  */
@@ -652,8 +656,9 @@ read_makefile (filename, type)
 
 	  /* Parse the dependencies.  */
 	  deps = (struct dep *)
-	    multi_glob (parse_file_seq (&p2, '\0', sizeof (struct dep), 1),
-			sizeof (struct dep));
+	    multi_glob (parse_file_seq (&p2, '\0', sizeof (struct dep)),
+			sizeof (struct dep),
+			1);
 
 	  commands_idx = 0;
 	  if (cmdleft != 0)
@@ -1415,13 +1420,16 @@ find_percent (pattern)
 
    SIZE is how big to construct chain elements.
    This is useful if we want them actually to be other structures
-   that have room for additional info.  */
+   that have room for additional info.
+
+   If STRIP is nonzero, strip `./'s off the beginning.  */
 
 struct nameseq *
-parse_file_seq (stringp, stopchar, size)
+parse_file_seq (stringp, stopchar, size, strip)
      char **stringp;
      char stopchar;
      unsigned int size;
+     int strip;
 {
   register struct nameseq *new = 0;
   register struct nameseq *new1;
@@ -1453,14 +1461,15 @@ parse_file_seq (stringp, stopchar, size)
 	}
       p--;
 
-      /* Skip leading `./'s.  */
-      while (p - q > 2 && q[0] == '.' && q[1] == '/')
-	{
-	  q += 2;		/* Skip "./".  */
-	  while (q < p && *q == '/')
-	    /* Skip following slashes: ".//foo" is "foo", not "/foo".  */
-	    ++q;
-	}
+      if (strip)
+	/* Skip leading `./'s.  */
+	while (p - q > 2 && q[0] == '.' && q[1] == '/')
+	  {
+	    q += 2;		/* Skip "./".  */
+	    while (q < p && *q == '/')
+	      /* Skip following slashes: ".//foo" is "foo", not "/foo".  */
+	      ++q;
+	  }
 
       /* Extract the filename just found, and skip it.  */
 
