@@ -355,19 +355,35 @@ define_automatic_variables ()
   /* This won't override any definition, but it
      will provide one if there isn't one there.  */
   v = define_variable ("SHELL", 5, default_shell, o_default, 0);
-  v->export = v_export;
+  v->export = v_export;		/* Always export SHELL.  */
 
   /* Don't let SHELL come from the environment.  */
   if (*v->value == '\0' || v->origin == o_env || v->origin == o_env_override)
     {
       free (v->value);
       v->origin = o_file;
-      v->value = savestring (default_shell, 7);
+      v->value = savestring (default_shell, strlen (default_shell));
     }
 
   /* Make sure MAKEFILES gets exported if it is set.  */
   v = define_variable ("MAKEFILES", 9, "", o_default, 0);
   v->export = v_ifset;
+
+  /* Define the magic D and F variables in terms of
+     the automatic variables they are variations of.  */
+
+  define_variable ("@D", 2, "$(dir $@)", o_automatic, 1);
+  define_variable ("%D", 2, "$(dir $%)", o_automatic, 1);
+  define_variable ("*D", 2, "$(dir $*)", o_automatic, 1);
+  define_variable ("<D", 2, "$(dir $<)", o_automatic, 1);
+  define_variable ("?D", 2, "$(dir $?)", o_automatic, 1);
+  define_variable ("^D", 2, "$(dir $^)", o_automatic, 1);
+  define_variable ("@F", 2, "$(notdir $@)", o_automatic, 1);
+  define_variable ("%F", 2, "$(notdir $%)", o_automatic, 1);
+  define_variable ("*F", 2, "$(notdir $*)", o_automatic, 1);
+  define_variable ("<F", 2, "$(notdir $<)", o_automatic, 1);
+  define_variable ("?F", 2, "$(notdir $?)", o_automatic, 1);
+  define_variable ("^F", 2, "$(notdir $^)", o_automatic, 1);
 }
 
 int export_all_variables;
@@ -437,6 +453,10 @@ target_environment (file)
 	      switch (v->export)
 		{
 		case v_default:
+		  if (v->origin == o_default || v->origin == o_automatic)
+		    /* Only export default variables by explicit request.  */
+		    continue;
+
 		  if (!export_all_variables
 		      && v->origin != o_command
 		      && v->origin != o_env && v->origin != o_env_override
@@ -490,11 +510,18 @@ target_environment (file)
       for (b = table[i]; b != 0; b = b->next)
 	{
 	  register struct variable *v = b->variable;
-	  /* If V is recursively expanded, expand its value.  */
-	  char *value = v->recursive ? recursively_expand (v) : v->value;
-	  result[nvariables++] = concat (v->name, "=", value);
-	  if (v->recursive)
-	    free (value);
+	  /* If V is recursively expanded and didn't come from the environment,
+	     expand its value.  If it came from the environment, it should
+	     go back into the environment unchanged.  */
+	  if (v->recursive
+	      && v->origin != o_env && v->origin != o_env_override)
+	    {
+	      char *value = recursively_expand (v);
+	      result[nvariables++] = concat (v->name, "=", value);
+	      free (value);
+	    }
+	  else
+	    result[nvariables++] = concat (v->name, "=", v->value);
 	}
     }
   result[nvariables] = (char *) xmalloc (100);
@@ -521,7 +548,9 @@ target_environment (file)
    If not, NULL is returned.  */
 
 struct variable *
-try_variable_definition (line, origin)
+try_variable_definition (filename, lineno, line, origin)
+     char *filename;
+     unsigned int lineno;
      char *line;
      enum variable_origin origin;
 {
@@ -529,8 +558,8 @@ try_variable_definition (line, origin)
   register char *p = line;
   register char *beg;
   register char *end;
-  register int recursive;
-  char *name, *expanded_name;
+  enum { bogus, simple, recursive, append } flavor = bogus;
+  char *name, *expanded_name, *value;
   struct variable *v;
 
   while (1)
@@ -540,36 +569,102 @@ try_variable_definition (line, origin)
 	return 0;
       if (c == '=')
 	{
-	  recursive = 1;
+	  end = p - 1;
+	  flavor = recursive;
 	  break;
 	}
       else if (c == ':')
 	if (*p == '=')
 	  {
-	    ++p;
-	    recursive = 0;
+	    end = p++ - 1;
+	    flavor = simple;
 	    break;
 	  }
 	else
+	  /* A colon other than := is a rule line, not a variable defn.  */
 	  return 0;
+      else if (c == '+' && *p == '=')
+	{
+	  end = p++ - 1;
+	  flavor = append;
+	  break;
+	}
     }
 
-  beg = next_token (line);
-  end = p - 1;
-  if (!recursive)
-    --end;
   while (isblank (end[-1]))
     --end;
+  beg = next_token (line);
   p = next_token (p);
 
   /* Expand the name, so "$(foo)bar = baz" works.  */
-  name = savestring (beg, end - beg);
+  name = (char *) alloca (end - beg + 1);
+  bcopy (beg, name, end - beg);
+  name[end - beg] = '\0';
   expanded_name = allocated_variable_expand (name);
-  free (name);
+
+  if (expanded_name[0] == '\0')
+    {
+      if (filename == 0)
+	fatal ("empty variable name");
+      else
+	makefile_fatal (filename, lineno, "empty variable name");
+    }
+
+  /* Calculate the variable's new value in VALUE.  */
+
+  switch (flavor)
+    {
+    case bogus:
+      /* Should not be possible.  */
+      abort ();
+      break;
+    case simple:
+      /* A simple variable definition "var := value".  Expand the value.  */
+      value = variable_expand (p);
+      break;
+    case recursive:
+      /* A recursive variable definition "var = value".
+	 The value is used verbatim.  */
+      value = p;
+      break;
+    case append:
+      /* An appending variable definition "var += value".
+	 Extract the old value and append the new one.  */
+      v = lookup_variable (expanded_name, strlen (expanded_name));
+      if (v == 0)
+	{
+	  /* There was no old value.
+	     This becomes a normal recursive definition.  */
+	  value = p;
+	  flavor = recursive;
+	}
+      else
+	{
+	  /* Paste the old and new values together in VALUE.  */
+
+	  unsigned int oldlen, newlen;
+
+	  if (v->recursive)
+	    /* The previous definition of the variable was recursive.
+	       The new value comes from the unexpanded old and new values.  */
+	    flavor = recursive;
+	  else
+	    /* The previous definition of the variable was simple.
+	       The new value comes from the old value, which was expanded
+	       when it was set; and from the expanded new value.  */
+	    p = variable_expand (p);
+
+	  oldlen = strlen (v->value);
+	  newlen = strlen (p);
+	  value = (char *) alloca (oldlen + 1 + newlen + 1);
+	  bcopy (v->value, value, oldlen);
+	  value[oldlen] = ' ';
+	  bcopy (p, &value[oldlen + 1], newlen + 1);
+	}
+    }
 
   v = define_variable (expanded_name, strlen (expanded_name),
-		       recursive ? p : variable_expand (p),
-		       origin, recursive);
+		       value, origin, flavor == recursive);
 
   free (expanded_name);
 
@@ -682,7 +777,7 @@ print_variable_set (set, prefix)
 #ifndef	NO_FLOAT
       printf ("# average of %.1f variables per bucket, \
 max %u in one bucket.\n",
-	      ((double) nvariables) * 100.0 / (double) set->buckets,
+	      (double) nvariables / (double) set->buckets,
 	      per_bucket);
 #endif
     }
